#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>

#include <runone.h>
#include <X11/Xlib.h>
#include <FL/Fl.H>
#include <FL/x.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Menu_Item.H>

#include "config.h"
#include "wmx.h"

void fork_exec(const char* name, const char* arg1, const char* arg2) {
	if (fork() == 0) {
		if (fork() == 0) {
			execl(name, name, arg1, arg2, 0);
			fprintf(stderr, "vrsm: execling %s: %s\n",
				name, strerror(errno));
			exit(1);
		}
		exit(0);
	}
	wait(0);
}

void on_screen_timeout(void* w) {
	reinterpret_cast<Fl_Window*>(w)->hide();
}

void on_screen(const char* s) {
	static Fl_Window* w;
	static Fl_Box* b;

	if(!w) {
		w = new Fl_Window((Fl::w() - 100) / 2, (Fl::h() - 9) * 2 / 3,
			100, 18);
		w->clear_border();
		w->color(FL_BLACK);
		b = new Fl_Box(0, 0, w->w(), w->h());
		b->labelsize(10);
		b->labelcolor(FL_WHITE);
		w->end();
	}

	b->label(s);
	b->redraw();
	w->show();

	Fl::flush();

	Fl::add_timeout((unsigned int)1000, on_screen_timeout, (void*)w);
}

void spawn_cb(Fl_Widget*, void*n) {
	char* name = (char*)n;
	//printf ("NAME: %s\n", name);
	// HACK: don't execsv xterm
	if(strstr(name, "Terminal") != NULL)
		fork_exec(name, 0, 0);
	else {
		char* argv[2] = {name, 0};
		runone(argv);
	}
}

bool check_client(Display* display, Window xwindow, Atom state_atom) {
	bool valid_client = true;

	if (valid_client) {
		Atom dummy1;
		int format = 0;
		unsigned long dummy2, nitems = 0, * prop = NULL;
      
		XGetWindowProperty (display, xwindow, state_atom,
			0, 1024, False, state_atom, &dummy1, &format,
			&nitems, &dummy2, (unsigned char **) &prop);
		if (prop) {
			valid_client = format == 32 && nitems > 0 &&
				(prop[0] == NormalState
					|| prop[0] == IconicState);
			XFree (prop);
		}
	}

	if (valid_client) {
		XWindowAttributes attributes = { 0, };

		XGetWindowAttributes (display, xwindow, &attributes);
		valid_client = (attributes.c_class == InputOutput &&
			attributes.map_state == IsViewable);
	}

	if (valid_client) {
		XWMHints* hints = XGetWMHints (display, xwindow);
      
		valid_client &= hints && (hints->flags & InputHint)
			&& hints->input;
		if (hints)
			XFree (hints);
	}
	
	return valid_client;
}

Window find_input_client(Display* display, Window xwindow, Atom state_atom) {
	if (check_client (display, xwindow, state_atom))
		return xwindow;

	Window dummy1, dummy2, * children = NULL;
	unsigned int n_children = 0;
	if (!XQueryTree (display, xwindow, &dummy1, &dummy2,
		&children, &n_children) || !children)
		return None;

	unsigned int i;
	for (i = 0; i < n_children; i++) {
		xwindow = find_input_client (display, children[i], state_atom);
		if (xwindow)
		break;
	}

	XFree(children);
 
	return xwindow;
}

static void raise_cb(Fl_Widget* men_item, void* data) {
	Window w = (Window)data;
	XRaiseWindow(fl_display, w);
}	

// use this to fill in a menu location:
static void init(Fl_Menu_Item& m, const char* data)
{
#ifdef HAVE_STYLES
	m.style = 0;
#endif
	m.label(data);
	m.flags = 0;
	m.labeltype(FL_NORMAL_LABEL);
	m.shortcut(0);
	m.labelfont(MENU_FONT_SLOT);
	m.labelsize(MENU_FONT_SIZE);
	m.labelcolor(FL_BLACK);
}

int add_window_submenu(Fl_Menu_Item* menu)
{
	// Retrieve information about currently open windows
	Window root = XRootWindow(fl_display, fl_screen);

	static Window* root_children = 0;
	if(root_children) {
		XFree(root_children);
		root_children = 0;
	}

	Window root_root;
	Window root_parent;
	unsigned int number_of_window;
	XQueryTree(fl_display, root, &root_root, &root_parent,
		&root_children, &number_of_window);

	// No items yet.
	int n = 0;

	// Add the 'Window' submenu
	init(menu[n], "Window");
	menu[n].flags = FL_SUBMENU | FL_MENU_DIVIDER;
	menu[n].callback((Fl_Callback*)0);
	n++;

	// Manage the storage for the window names.
	// HACK: this is a nasty hack.  make this dynamic.
	static char** winlist[64];
	int i;
	for(i = 0; i < 64; i++)
		if(winlist[i]) {
			XFreeStringList(winlist[i]);
			winlist[i] = 0;
		}

	// Add the window list
	for(int i = 0; i < number_of_window; i++) {
		Window xwindow = find_input_client(fl_display, root_children[i],
			XInternAtom(fl_display, "WM_STATE", False));
		if(!xwindow)
			continue;

		XTextProperty text;
		XGetWMName(fl_display, xwindow, &text);	
		int nsb;
		char** text_list;
		if(!XTextPropertyToStringList(&text, &text_list, &nsb))
			continue;
		if(nsb <= 0)
			continue;

		winlist[n] = text_list;
		init(menu[n], winlist[n][0]);
		menu[n].callback(raise_cb, (void*)xwindow);

		n++;
	}

	// Make the Window menu inactive if no windows are on the list
	if(n == 1)
		menu[0].flags |= FL_MENU_INACTIVE;

	// Close off 'Window' submenu
	init(menu[n], 0);
	n++;

	return n;
}

int add_wmx_items(Fl_Menu_Item* menu)
{
	static char** wmxlist;
	static int wmxlist_capacity;
	static int wmx_pathlen;
	static int wmxlist_items;
	load_wmx(&wmxlist, wmxlist_capacity, &wmx_pathlen, &wmxlist_items);

	// Did the buffer grow?
	if(wmxlist_items > wmxlist_capacity)
		wmxlist_capacity = wmxlist_items;

	// Add wmx items to the menu.
	int n = 0;
	int level = 0;
	int pathlen[MAX_NESTING_DEPTH];
	pathlen[0] = wmx_pathlen;
	for(int i = 0; i < wmxlist_items; i++) {
		char* cmd = wmxlist[i];
		cmd += strspn(cmd, "/") - 1;
		init(menu[n], cmd + pathlen[level]);

		int nextlev = (i == wmxlist_items - 1) ?
			0 : strspn(wmxlist[i + 1], "/") - 1;
		if (nextlev < level) {
			// Last item on a submenu
			menu[n].callback(spawn_cb, cmd);
			// Close 'em off
			for (; level>nextlev; level--)
				init(menu[++n], 0);
		} else if (nextlev > level) {
			// New submenu
			// This should be made a submenu
 			// extra for next trailing /
			pathlen[++level] = strlen(cmd) + 1;
			menu[n].flags = FL_SUBMENU;
			menu[n].callback((Fl_Callback*)0);
		} else {
			// Selectable menu item
			menu[n].callback(spawn_cb, cmd);
		}
		n++;
	}

	return(n);
}

extern int orientation;
void show_menu()
{
	// Fixed size for now.  Old code tried to be dynamic
	// and was quite broken.  bdl
	static Fl_Menu_Item* menu;
	if(!menu)
		menu = new Fl_Menu_Item[256];

	// No menu items yet.
	int n = 0;

	// Add the "Window" submenu to the menu.
	n += add_window_submenu(&menu[n]);

	// Add the wmx items to the menu.
	n += add_wmx_items(&menu[n]);

	// Terminate the menu item array.
	menu[n].label(0);

	// Draw an FL_MENU_DIVIDER after the submenus.
	int i = n - 1; 
	while((i >= 0) && (menu[i].flags != FL_SUBMENU))
		i--;
	if(i >= 0)
		menu[i].flags |= FL_MENU_DIVIDER;

	// Do the menu.
	const Fl_Menu_Item* picked;
	if (orientation == 0) 
	  picked = menu->pulldown(0, -4, 104, 44);
	else if (orientation == 270)
	  picked = menu->popup(-100, 160);
	else if (orientation == 90)
	  picked = menu->popup(27, 120);
	else // if (orientation == 180)
	  picked = menu->popup(60, 27);

	if(picked && picked->callback()) {
		on_screen(picked->label());
		picked->do_callback(0);
	}
}
